"""
=======
Network
=======

Network visualization and analysis tools for Orange.

"""

# Category description for the widget registry

NAME = "Network"

DESCRIPTION = "Network visualization and analysis tools for Orange."

BACKGROUND = "#C0FF97"

ICON = "icons/Category-Network.svg"
